/*
 * kgraft_patch_bsc960563
 *
 * Fix for CVE-2015-8709, bsc#960563
 *
 *  SLE12 commit:
 *  aefdacd669c92b64e5d9ed94331886cfa811803d
 *
 *  Copyright (c) 2016 SUSE
 *  Author: Miroslav Benes <mbenes@suse.cz>
 *
 *  Based on the original Linux kernel code. Other copyrights apply.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/kallsyms.h>
#include <linux/ptrace.h>
#include <linux/user_namespace.h>

static int (*kgr_security_ptrace_access_check)(struct task_struct *child, unsigned int mode);
static bool (*kgr_has_ns_capability_noaudit)(struct task_struct *t, struct user_namespace *ns, int cap);
static bool (*kgr_has_ns_capability)(struct task_struct *t, struct user_namespace *ns, int cap);
static int (*kgr_get_dumpable)(struct mm_struct *mm);

static struct {
	char *name;
	void **addr;
} kgr_funcs[] = {
	{ "security_ptrace_access_check", (void *)&kgr_security_ptrace_access_check },
	{ "has_ns_capability_noaudit", (void *)&kgr_has_ns_capability_noaudit },
	{ "has_ns_capability", (void *)&kgr_has_ns_capability },
	{ "get_dumpable", (void *)&kgr_get_dumpable },
};

/*
 * Fix CVE-2015-8709
 *
 * Changed and patched function
 * -static int ptrace_has_cap(struct user_namespace *ns, unsigned int mode)
 */
bool kgr_ptrace_has_cap(const struct cred *tcred, unsigned int mode)
{
	/*
	 * Fix CVE-2015-8709
	 *  +next 20 rows
	 */
	struct user_namespace *tns = tcred->user_ns;

	/* When a root-owned process enters a user namespace created by a
	 * malicious user, the user shouldn't be able to execute code under
	 * uid 0 by attaching to the root-owned process via ptrace.
	 * Therefore, similar to the capable_wrt_inode_uidgid() check,
	 * verify that all the uids and gids of the target process are
	 * mapped into a namespace below the current one in which the caller
	 * is capable.
	 * No fsuid/fsgid check because __ptrace_may_access doesn't do it
	 * either.
	 */
	while (
	    !kuid_has_mapping(tns, tcred->euid) ||
	    !kuid_has_mapping(tns, tcred->suid) ||
	    !kuid_has_mapping(tns, tcred->uid)  ||
	    !kgid_has_mapping(tns, tcred->egid) ||
	    !kgid_has_mapping(tns, tcred->sgid) ||
	    !kgid_has_mapping(tns, tcred->gid)) {
		tns = tns->parent;
	}

	if (mode & PTRACE_MODE_NOAUDIT)
		/*
		 * Fix CVE-2015-8709
		 *  -return has_ns_capability_noaudit(current, ns, CAP_SYS_PTRACE);
		 *  +next row
		 */
		return kgr_has_ns_capability_noaudit(current, tns, CAP_SYS_PTRACE);
	else
		/*
		 * Fix CVE-2015-8709
		 *  -return has_ns_capability(current, ns, CAP_SYS_PTRACE);
		 *  +next row
		 */
		return kgr_has_ns_capability(current, tns, CAP_SYS_PTRACE);
}

/* Patched function */
int kgr__ptrace_may_access(struct task_struct *task, unsigned int mode)
{
	const struct cred *cred = current_cred(), *tcred;

	/* May we inspect the given task?
	 * This check is used both for attaching with ptrace
	 * and for allowing access to sensitive information in /proc.
	 *
	 * ptrace_attach denies several cases that /proc allows
	 * because setting up the necessary parent/child relationship
	 * or halting the specified task is impossible.
	 */
	int dumpable = 0;
	/* Don't let security modules deny introspection */
	if (same_thread_group(task, current))
		return 0;
	rcu_read_lock();
	tcred = __task_cred(task);
	if (uid_eq(cred->uid, tcred->euid) &&
	    uid_eq(cred->uid, tcred->suid) &&
	    uid_eq(cred->uid, tcred->uid)  &&
	    gid_eq(cred->gid, tcred->egid) &&
	    gid_eq(cred->gid, tcred->sgid) &&
	    gid_eq(cred->gid, tcred->gid))
		goto ok;
	/*
	 * Fix CVE-2015-8709
	 *  -if (ptrace_has_cap(tcred->user_ns, mode))
	 *  +next row
	 */
	if (kgr_ptrace_has_cap(tcred, mode))
		goto ok;
	rcu_read_unlock();
	return -EPERM;
ok:
	rcu_read_unlock();
	smp_rmb();
	if (task->mm)
		dumpable = kgr_get_dumpable(task->mm);
	rcu_read_lock();
	if (dumpable != SUID_DUMP_USER &&
	    /*
	     * Fix CVE-2015-8709
	     *  -!ptrace_has_cap(__task_cred(task)->user_ns, mode)) {
	     */
	    !kgr_ptrace_has_cap(__task_cred(task), mode)) {
		rcu_read_unlock();
		return -EPERM;
	}
	rcu_read_unlock();

	return kgr_security_ptrace_access_check(task, mode);
}

static int kgr_patch_bsc960563_kallsyms(void)
{
	unsigned long addr;
	int i;

	for (i = 0; i < ARRAY_SIZE(kgr_funcs); i++) {
		addr = kallsyms_lookup_name(kgr_funcs[i].name);
		if (!addr) {
			pr_err("kgraft-patch: symbol %s not resolved\n",
				kgr_funcs[i].name);
			return -ENOENT;
		}

		*(kgr_funcs[i].addr) = (void *)addr;
	}

	return 0;
}

int kgr_patch_bsc960563_init(void)
{
	return kgr_patch_bsc960563_kallsyms();
}
